@@ -1,7 +1,14 @@
 require "benchmark/ips"
 require "json"
-require "oj"
-require "rapidjson"
+begin
+  require "oj"
+rescue LoadError
+end
+
+begin
+  require "rapidjson"
+rescue LoadError
+end
 
 if ENV["ONLY"]
   RUN = ENV["ONLY"].split(/[,: ]/).map{|x| [x.to_sym, true] }.to_h
@@ -15,12 +22,21 @@ end
 
 def benchmark_parsing(name, json_output)
   puts "== Parsing #{name} (#{json_output.size} bytes)"
+  coder = JSON::Coder.new
 
   Benchmark.ips do |x|
     x.report("json")      { JSON.parse(json_output) } if RUN[:json]
-    x.report("oj")        { Oj.load(json_output) } if RUN[:oj]
-    x.report("Oj::Parser") { Oj::Parser.new(:usual).parse(json_output) } if RUN[:oj]
-    x.report("rapidjson") { RapidJSON.parse(json_output) } if RUN[:rapidjson]
+    x.report("json_coder") { coder.load(json_output) } if RUN[:json_coder]
+
+    if defined?(Oj)
+      x.report("oj")        { Oj.load(json_output) } if RUN[:oj]
+      x.report("Oj::Parser") { Oj::Parser.new(:usual).parse(json_output) } if RUN[:oj]
+    end
+
+    if defined?(RapidJSON)
+      x.report("rapidjson") { RapidJSON.parse(json_output) } if RUN[:rapidjson]
+    end
+
     x.compare!(order: :baseline)
   end
   puts
@@ -41,9 +57,4 @@ JSON
 benchmark_parsing "activitypub.json", File.read("#{__dir__}/data/activitypub.json")
 benchmark_parsing "twitter.json", File.read("#{__dir__}/data/twitter.json")
 benchmark_parsing "citm_catalog.json", File.read("#{__dir__}/data/citm_catalog.json")
-
-# rapidjson is 8x faster thanks to its much more performant float parser.
-# Unfortunately, there isn't a lot of existing fast float parsers in pure C,
-# and including C++ is problematic.
-# Aside from that, we're close to the alternatives here.
 benchmark_parsing "float parsing", File.read("#{__dir__}/data/canada.json")
