@@ -14,6 +14,7 @@ import org.jruby.RubyHash;
 import org.jruby.RubyInteger;
 import org.jruby.RubyNumeric;
 import org.jruby.RubyObject;
+import org.jruby.RubyProc;
 import org.jruby.RubyString;
 import org.jruby.anno.JRubyMethod;
 import org.jruby.runtime.Block;
@@ -22,6 +23,7 @@ import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.Visibility;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.util.ByteList;
+import org.jruby.util.TypeConverter;
 
 /**
  * The <code>JSON::Ext::Generator::State</code> class.
@@ -32,6 +34,9 @@ import org.jruby.util.ByteList;
  * @author mernen
  */
 public class GeneratorState extends RubyObject {
+    private boolean allowDuplicateKey = false;
+    private boolean deprecateDuplicateKey = true;
+
     /**
      * The indenting unit string. Will be repeated several times for larger
      * indenting levels.
@@ -58,6 +63,8 @@ public class GeneratorState extends RubyObject {
      */
     private ByteList arrayNl = ByteList.EMPTY_BYTELIST;
 
+    private RubyProc asJSON;
+
     /**
      * The maximum level of nesting of structures allowed.
      * <code>0</code> means disabled.
@@ -78,12 +85,6 @@ public class GeneratorState extends RubyObject {
      */
     private boolean asciiOnly = DEFAULT_ASCII_ONLY;
     static final boolean DEFAULT_ASCII_ONLY = false;
-    /**
-     * If set to <code>true</code> all JSON values generated might not be
-     * RFC-conform JSON documents.
-     */
-    private boolean quirksMode = DEFAULT_QUIRKS_MODE;
-    static final boolean DEFAULT_QUIRKS_MODE = false;
     /**
      * If set to <code>true</code> the forward slash will be escaped in
      * json output.
@@ -106,7 +107,7 @@ public class GeneratorState extends RubyObject {
     /**
      * The current depth (inside a #to_json call)
      */
-    private int depth = 0;
+    protected int depth = 0;
 
     static final ObjectAllocator ALLOCATOR = GeneratorState::new;
 
@@ -134,7 +135,7 @@ public class GeneratorState extends RubyObject {
 
     @JRubyMethod(meta=true)
     public static IRubyObject generate(ThreadContext context, IRubyObject klass, IRubyObject obj, IRubyObject opts, IRubyObject io) {
-        return fromState(context, opts)._generate(context, obj, io);
+        return fromState(context, opts).generate(context, obj, io);
     }
 
     static GeneratorState fromState(ThreadContext context, IRubyObject opts) {
@@ -154,8 +155,7 @@ public class GeneratorState extends RubyObject {
             }
         }
 
-        // for other values, return the safe prototype
-        return (GeneratorState)info.getSafeStatePrototype(context).dup();
+        return (GeneratorState)klass.newInstance(context, context.nil);
     }
 
     /**
@@ -211,14 +211,18 @@ public class GeneratorState extends RubyObject {
         this.spaceBefore = orig.spaceBefore;
         this.objectNl = orig.objectNl;
         this.arrayNl = orig.arrayNl;
+        this.asJSON = orig.asJSON;
         this.maxNesting = orig.maxNesting;
         this.allowNaN = orig.allowNaN;
         this.asciiOnly = orig.asciiOnly;
-        this.quirksMode = orig.quirksMode;
         this.scriptSafe = orig.scriptSafe;
         this.strict = orig.strict;
         this.bufferInitialLength = orig.bufferInitialLength;
         this.depth = orig.depth;
+
+        this.allowDuplicateKey = orig.allowDuplicateKey;
+        this.deprecateDuplicateKey = orig.deprecateDuplicateKey;
+
         return this;
     }
 
@@ -227,8 +231,8 @@ public class GeneratorState extends RubyObject {
      * the result. If no valid JSON document can be created this method raises
      * a GeneratorError exception.
      */
-    @JRubyMethod(visibility = Visibility.PRIVATE)
-    public IRubyObject _generate(ThreadContext context, IRubyObject obj, IRubyObject io) {
+    @JRubyMethod
+    public IRubyObject generate(ThreadContext context, IRubyObject obj, IRubyObject io) {
         IRubyObject result = Generator.generateJson(context, obj, this, io);
         RuntimeInfo info = RuntimeInfo.forRuntime(context.runtime);
         if (!(result instanceof RubyString)) {
@@ -247,6 +251,11 @@ public class GeneratorState extends RubyObject {
         return resultString;
     }
 
+    @JRubyMethod
+    public IRubyObject generate(ThreadContext context, IRubyObject obj) {
+        return generate(context, obj, context.nil);
+    }
+
     @JRubyMethod(name="[]")
     public IRubyObject op_aref(ThreadContext context, IRubyObject vName) {
         String name = vName.asJavaString();
@@ -260,6 +269,7 @@ public class GeneratorState extends RubyObject {
 
     @JRubyMethod(name="[]=")
     public IRubyObject op_aset(ThreadContext context, IRubyObject vName, IRubyObject value) {
+        checkFrozen();
         String name = vName.asJavaString();
         String nameWriter = name + "=";
         if (getMetaClass().isMethodBound(nameWriter, true)) {
@@ -281,6 +291,7 @@ public class GeneratorState extends RubyObject {
 
     @JRubyMethod(name="indent=")
     public IRubyObject indent_set(ThreadContext context, IRubyObject indent) {
+        checkFrozen();
         this.indent = prepareByteList(context, indent);
         return indent;
     }
@@ -296,6 +307,7 @@ public class GeneratorState extends RubyObject {
 
     @JRubyMethod(name="space=")
     public IRubyObject space_set(ThreadContext context, IRubyObject space) {
+        checkFrozen();
         this.space = prepareByteList(context, space);
         return space;
     }
@@ -312,6 +324,7 @@ public class GeneratorState extends RubyObject {
     @JRubyMethod(name="space_before=")
     public IRubyObject space_before_set(ThreadContext context,
                                         IRubyObject spaceBefore) {
+        checkFrozen();
         this.spaceBefore = prepareByteList(context, spaceBefore);
         return spaceBefore;
     }
@@ -328,6 +341,7 @@ public class GeneratorState extends RubyObject {
     @JRubyMethod(name="object_nl=")
     public IRubyObject object_nl_set(ThreadContext context,
                                      IRubyObject objectNl) {
+        checkFrozen();
         this.objectNl = prepareByteList(context, objectNl);
         return objectNl;
     }
@@ -344,10 +358,31 @@ public class GeneratorState extends RubyObject {
     @JRubyMethod(name="array_nl=")
     public IRubyObject array_nl_set(ThreadContext context,
                                     IRubyObject arrayNl) {
+        checkFrozen();
         this.arrayNl = prepareByteList(context, arrayNl);
         return arrayNl;
     }
 
+    public RubyProc getAsJSON() {
+        return asJSON;
+    }
+
+    @JRubyMethod(name="as_json")
+    public IRubyObject as_json_get(ThreadContext context) {
+        return asJSON == null ? context.getRuntime().getFalse() : asJSON;
+    }
+
+    @JRubyMethod(name="as_json=")
+    public IRubyObject as_json_set(ThreadContext context, IRubyObject asJSON) {
+        checkFrozen();
+        if (asJSON.isNil() || asJSON == context.getRuntime().getFalse()) {
+            this.asJSON = null;
+        } else {
+            this.asJSON = (RubyProc)TypeConverter.convertToType(asJSON, context.getRuntime().getProc(), "to_proc");
+        }
+        return asJSON;
+    }
+
     @JRubyMethod(name="check_circular?")
     public RubyBoolean check_circular_p(ThreadContext context) {
         return RubyBoolean.newBoolean(context, maxNesting != 0);
@@ -360,6 +395,7 @@ public class GeneratorState extends RubyObject {
 
     @JRubyMethod(name="max_nesting=")
     public IRubyObject max_nesting_set(IRubyObject max_nesting) {
+        checkFrozen();
         maxNesting = RubyNumeric.fix2int(max_nesting);
         return max_nesting;
     }
@@ -378,6 +414,7 @@ public class GeneratorState extends RubyObject {
 
     @JRubyMethod(name="script_safe=", alias="escape_slash=")
     public IRubyObject script_safe_set(IRubyObject script_safe) {
+        checkFrozen();
         scriptSafe = script_safe.isTrue();
         return script_safe.getRuntime().newBoolean(scriptSafe);
     }
@@ -401,6 +438,7 @@ public class GeneratorState extends RubyObject {
 
     @JRubyMethod(name="strict=")
     public IRubyObject strict_set(IRubyObject isStrict) {
+        checkFrozen();
         strict = isStrict.isTrue();
         return isStrict.getRuntime().newBoolean(strict);
     }
@@ -430,6 +468,7 @@ public class GeneratorState extends RubyObject {
 
     @JRubyMethod(name="buffer_initial_length=")
     public IRubyObject buffer_initial_length_set(IRubyObject buffer_initial_length) {
+        checkFrozen();
         int newLength = RubyNumeric.fix2int(buffer_initial_length);
         if (newLength > 0) bufferInitialLength = newLength;
         return buffer_initial_length;
@@ -446,6 +485,7 @@ public class GeneratorState extends RubyObject {
 
     @JRubyMethod(name="depth=")
     public IRubyObject depth_set(IRubyObject vDepth) {
+        checkFrozen();
         depth = RubyNumeric.fix2int(vDepth);
         return vDepth;
     }
@@ -458,6 +498,24 @@ public class GeneratorState extends RubyObject {
         return str.getByteList().dup();
     }
 
+    @JRubyMethod(name="allow_duplicate_key?", visibility=Visibility.PRIVATE)
+    public IRubyObject allow_duplicate_key_p(ThreadContext context) {
+        if (allowDuplicateKey) {
+            return context.runtime.getTrue();
+        } else if (deprecateDuplicateKey) {
+            return context.runtime.getNil();
+        }
+        return context.runtime.getFalse();
+    }
+
+    public boolean getAllowDuplicateKey() {
+        return allowDuplicateKey;
+    }
+
+    public boolean getDeprecateDuplicateKey() {
+        return deprecateDuplicateKey;
+    }
+
     /**
      * <code>State#configure(opts)</code>
      *
@@ -468,6 +526,7 @@ public class GeneratorState extends RubyObject {
      */
   @JRubyMethod(visibility=Visibility.PRIVATE)
     public IRubyObject _configure(ThreadContext context, IRubyObject vOpts) {
+        checkFrozen();
         OptionsReader opts = new OptionsReader(context, vOpts);
 
         ByteList indent = opts.getString("indent");
@@ -482,6 +541,8 @@ public class GeneratorState extends RubyObject {
         ByteList arrayNl = opts.getString("array_nl");
         if (arrayNl != null) this.arrayNl = arrayNl;
 
+        this.asJSON = opts.getProc("as_json");
+
         ByteList objectNl = opts.getString("object_nl");
         if (objectNl != null) this.objectNl = objectNl;
 
@@ -497,6 +558,10 @@ public class GeneratorState extends RubyObject {
 
         depth = opts.getInt("depth", 0);
 
+        if (opts.hasKey("allow_duplicate_key")) {
+            this.allowDuplicateKey = opts.getBool("allow_duplicate_key", false);
+            this.deprecateDuplicateKey = false;
+        }
         return this;
     }
 
@@ -517,6 +582,7 @@ public class GeneratorState extends RubyObject {
         result.op_aset(context, runtime.newSymbol("space_before"), space_before_get(context));
         result.op_aset(context, runtime.newSymbol("object_nl"), object_nl_get(context));
         result.op_aset(context, runtime.newSymbol("array_nl"), array_nl_get(context));
+        result.op_aset(context, runtime.newSymbol("as_json"), as_json_get(context));
         result.op_aset(context, runtime.newSymbol("allow_nan"), allow_nan_p(context));
         result.op_aset(context, runtime.newSymbol("ascii_only"), ascii_only_p(context));
         result.op_aset(context, runtime.newSymbol("max_nesting"), max_nesting_get(context));
@@ -524,6 +590,16 @@ public class GeneratorState extends RubyObject {
         result.op_aset(context, runtime.newSymbol("strict"), strict_get(context));
         result.op_aset(context, runtime.newSymbol("depth"), depth_get(context));
         result.op_aset(context, runtime.newSymbol("buffer_initial_length"), buffer_initial_length_get(context));
+
+        if (this.allowDuplicateKey) {
+            if (!this.deprecateDuplicateKey) {
+                result.op_aset(context, runtime.newSymbol("allow_duplicate_key"), runtime.getTrue());
+            }
+        }
+        else {
+            result.op_aset(context, runtime.newSymbol("allow_duplicate_key"), runtime.getFalse());
+        }
+
         for (String name: getInstanceVariableNameList()) {
             result.op_aset(context, runtime.newSymbol(name.substring(1)), getInstanceVariables().getInstanceVariable(name));
         }
@@ -536,8 +612,8 @@ public class GeneratorState extends RubyObject {
         return depth;
     }
 
-    public void decreaseDepth() {
-        --depth;
+    public int decreaseDepth() {
+        return --depth;
     }
 
     /**
@@ -547,7 +623,7 @@ public class GeneratorState extends RubyObject {
     private void checkMaxNesting(ThreadContext context) {
         if (maxNesting != 0 && depth > maxNesting) {
             depth--;
-            throw Utils.newException(context, Utils.M_NESTING_ERROR, "nesting of " + depth + " is too deep");
+            throw Utils.newException(context, Utils.M_NESTING_ERROR, "nesting of " + depth + " is too deep. Did you try to serialize objects with circular references?");
         }
     }
 }
